import datetime
import uuid


def create_deployment(environment, user, description):
    return {
        "environment": environment,
        "date_created": datetime.datetime.utcnow().isoformat(),
        "requested_by": user,
        "description": description
    }


def create_release(project_id, project_name, current_user, release_description, release_images):
    release_id = str(uuid.uuid4())
    return {
        "release_id": release_id,
        "project_id": project_id,
        "project_name": project_name,
        "date_created": datetime.datetime.utcnow().isoformat(),
        "requested_by": current_user,
        "description": release_description,
        "images": release_images,
        "deployments": []
    }



# -*- encoding: utf-8 -*-
"""Command running functions"""

import shlex
import subprocess
import sys


def cmd(*args):
    return subprocess.check_output(list(args)).decode("utf8").strip()


def ensure(command):
    try:
        subprocess.check_call(shlex.split(command))
    except subprocess.CalledProcessError as err:
        sys.exit(err.returncode)


def configure_aws_profile(role_arn, profile_name):
    cmd('aws', 'configure', 'set', 'region', "eu-west-1", '--profile', profile_name)  # noqa: E501
    cmd('aws', 'configure', 'set', 'role_arn', role_arn, '--profile', profile_name)   # noqa: E501
    cmd('aws', 'configure', 'set', 'source_profile', 'default', '--profile', profile_name)  # noqa: E501




def pprint_nested_tree(tree, indent=0):
    lines = []

    if indent == 0:
        lines.append(".")

    if isinstance(tree, str):
        return tree

    # This is a bit of a hard-coded fudge for a common case that causes it
    # to be printed in a slightly more useful way.
    if tree.keys() == {"latest", "prod", "stage"}:
        entries = [
            ("latest", tree["latest"]),
            ("stage", tree["stage"]),
            ("prod", tree["prod"])
        ]
    else:
        entries = sorted(tree.items())

    for i, (key, nested_tree) in enumerate(entries, start=1):
        if i == len(entries):
            lines.append("└── " + key)

            if isinstance(nested_tree, str):
                lines[-1] = lines[-1].ljust(40) + nested_tree
            else:
                lines.extend([
                    "    " + line
                    for line in pprint_nested_tree(nested_tree, indent=indent + 1)
                ])
        else:
            lines.append("├── " + key)
            if isinstance(nested_tree, str):
                lines[-1] = lines[-1].ljust(40) + nested_tree
            else:
                lines.extend([
                    "│   " + line
                    for line in pprint_nested_tree(nested_tree, indent=indent + 1)
                ])

    return lines


def build_tree_from_paths(paths):
    tree = {}

    for path, value in paths.items():
        curr_tree = tree
        path_components = path.strip("/").split("/")
        for component in path_components:
            try:
                curr_tree = curr_tree[component]
            except KeyError:
                if component == path_components[-1]:
                    curr_tree[component] = value
                else:
                    curr_tree[component] = {}
                curr_tree = curr_tree[component]

    return tree


def pprint_path_keyval_dict(paths):
    tree = build_tree_from_paths(paths)
    return pprint_nested_tree(tree)



import os

from botocore.exceptions import ClientError

from .iam import Iam
from .commands import cmd, ensure


class Ecr:
    def __init__(self, account_id, region_id, role_arn=None):
        self.account_id = account_id
        self.region_id = region_id
        self.session = Iam.get_session("ReleaseToolEcr", role_arn)
        self.ecr = self.session.client('ecr')

    @staticmethod
    def _get_release_image_tag(service_id):
        repo_root = cmd("git", "rev-parse", "--show-toplevel")
        release_file = os.path.join(repo_root, ".releases", service_id)

        print(f"*** Retrieving image tag for {service_id} from {release_file}")

        return open(release_file).read().strip()

    def publish_image(self, namespace, service_id, label, dry_run=False):
        # some terminology as label & tag are confusing
        # - label is a given label relating to deployment e.g. latest, prod, stage
        # - image_tag is usually the git ref, denoting a particular build artifact

        image_tag = Ecr._get_release_image_tag(service_id)

        label_image_name = f"{service_id}:{label}"
        tag_image_name = f"{service_id}:{image_tag}"

        base_remote_image_name = (
            f"{self.account_id}.dkr.ecr.{self.region_id}.amazonaws.com/{namespace}"
        )

        remote_label_image_name = f"{base_remote_image_name}/{label_image_name}"
        remote_tag_image_name = f"{base_remote_image_name}/{tag_image_name}"

        if not dry_run:
            try:
                print(f"*** Pushing {label_image_name} to {remote_label_image_name}")
                cmd('docker', 'tag', label_image_name, remote_label_image_name)
                cmd('docker', 'push', remote_label_image_name)

                print(f"*** Pushing {tag_image_name} to {remote_tag_image_name}")
                cmd('docker', 'tag', tag_image_name, remote_tag_image_name)
                cmd('docker', 'push', remote_tag_image_name)

            finally:
                cmd('docker', 'rmi', label_image_name)
                cmd('docker', 'rmi', tag_image_name)

        return remote_tag_image_name

    def retag_image(self, namespace, service_id, tag, new_tag, dry_run=False):
        repository_name = f"{namespace}/{service_id}"

        result = self.ecr.batch_get_image(
            registryId=self.account_id,
            repositoryName=repository_name,
            imageIds=[
                {"imageTag": tag}
            ]
        )

        if len(result["images"]) == 0:
            raise RuntimeError(f"No matching images found for {repository_name}:{tag}!")

        if len(result["images"]) > 1:
            raise RuntimeError(f"Multiple matching images found for {repository_name}:{tag}!")

        image = result["images"][0]

        if not dry_run:
            try:
                self.ecr.put_image(
                    registryId=self.account_id,
                    repositoryName=repository_name,
                    imageTag=new_tag,
                    imageManifest=image['imageManifest']
                )
            except ClientError as e:
                # Matching tag & digest already exists (nothing to do)
                if not e.response['Error']['Code'] == 'ImageAlreadyExistsException':
                    raise e

    def login(self, profile_name=None):
        base = ['aws', 'ecr', 'get-login']
        login_options = ['--no-include-email', '--registry-ids', self.account_id]
        profile_options = ['--profile', profile_name]

        if profile_name:
            login = base + profile_options + login_options
        else:
            login = base + login_options

        ensure(cmd(*login))



import boto3


class Iam:
    def __init__(self, role_arn=None):
        self.session = Iam.get_session("ReleaseToolIamUserDetails", role_arn)
        self.iam = self.session.resource('iam')

    @staticmethod
    def get_session(session_name, role_arn=None):
        if role_arn:
            client = boto3.client('sts')
            response = client.assume_role(
                RoleArn=role_arn,
                RoleSessionName=session_name
            )

            return boto3.session.Session(
                aws_access_key_id=response['Credentials']['AccessKeyId'],
                aws_secret_access_key=response['Credentials']['SecretAccessKey'],
                aws_session_token=response['Credentials']['SessionToken']
            )
        else:
            return boto3.session.Session()

    def caller_identity(self, underlying=False):
        if underlying:
            client = boto3.client('sts')
        else:
            client = self.session.client('sts')

        caller_identity = client.get_caller_identity()

        return {
            "arn": caller_identity['Arn'],
            "user_id": caller_identity['UserId'],
            "account_id": caller_identity['Account']
        }



# -*- encoding: utf-8

import json
import os


def load(project_filepath):
    with open(project_filepath) as infile:
        return json.load(infile)


def save(project_filepath, project):
    with open(project_filepath, 'w') as outfile:
        json.dump(project, outfile, sort_keys=True, indent=2)


def exists(project_filepath):
    return os.path.isfile(project_filepath)


def get_environments_lookup(project):
    return {
        e['id']: e for e in project['environments'] if 'id' in e
    }



import click
import json

from dateutil.parser import parse
from urllib.parse import urlparse
from pprint import pprint

from .commands import configure_aws_profile
from .ecr import Ecr
from .model import create_deployment, create_release
from .project_config import load, save, exists, get_environments_lookup

from .releases_store import DynamoDbReleaseStore
from .parameter_store import SsmParameterStore
from .pretty_printing import pprint_path_keyval_dict
from .iam import Iam

DEFAULT_PROJECT_FILEPATH = ".wellcome_project"
DEFAULT_ECR_NAMESPACE = "uk.ac.wellcome"


def _format_ecr_uri(uri):
    image_name = uri.split("/")[2]
    image_label, image_tag = image_name.split(":")

    return {
        'label': image_label,
        'tag': image_tag[:7]
    }


def _summarise_ssm_response(images):
    for image in images:
        yield {
            'name': image['Name'],
            'value': image['Value'],
            'last_modified': image['LastModifiedDate'].strftime('%d-%m-%YT%H:%M')
        }


def _summarise_release_deployments(releases):
    summaries = []
    for r in releases:
        for d in r['deployments']:
            summaries.append(
                {
                    'release_id': r['release_id'][:8],
                    'environment_id': d['environment']['id'],
                    'deployed_date': parse(d['date_created']).strftime('%d-%m-%YT%H:%M'),
                    'description': d['description']
                }
            )
    return summaries


def _is_url(label):
    try:
        res = urlparse(label)
        return all([res.scheme, res.netloc])
    except ValueError:
        return False


@click.group()
@click.option('--project-file', '-f', default=DEFAULT_PROJECT_FILEPATH)
@click.option('--verbose', '-v', is_flag=True, help="Print verbose messages.")
@click.option("--project-id", '-i', help="Specify the project ID")
@click.option("--region-id", '-i', help="Specify the AWS region ID")
@click.option("--account-id", help="Specify the AWS account ID")
@click.option("--role-arn")
@click.option('--dry-run', '-d', is_flag=True, help="Don't make changes.")
@click.pass_context
def cli(ctx, project_file, verbose, project_id, region_id, account_id, role_arn, dry_run):
    try:
        projects = load(project_file)
    except FileNotFoundError:
        if ctx.invoked_subcommand != "initialise":
            message = f"Couldn't find project metadata file {project_file!r}.  Run `initialise`."
            raise click.UsageError(message) from None
        ctx.obj = {
            'project_filepath': project_file,
            'verbose': verbose,
            'dry_run': dry_run,
        }
        return

    project_names = list(projects.keys())
    project_count = len(project_names)

    if not project_id:
        if project_count == 1:
            project_id = project_names[0]
        else:
            project_id = click.prompt(
                text="Enter the project ID",
                type=click.Choice(project_names)
            )

    project = projects.get(project_id)
    project['id'] = project_id

    if verbose and project:
        click.echo(f"Loaded {project_file} {project}")

    if role_arn:
        project['role_arn'] = role_arn
        if verbose:
            click.echo(f"Using role_arn {project['role_arn']}")

    if region_id:
        project['aws_region_name'] = region_id
        if verbose:
            click.echo(f"Using aws_region_name {project['aws_region_name']}")

    user_details = Iam(role_arn)
    caller_identity = user_details.caller_identity()
    underlying_caller_identity = user_details.caller_identity(underlying=True)

    if verbose:
        click.echo(f"Running as role {caller_identity['arn']}")
        if caller_identity['arn'] != underlying_caller_identity['arn']:
            click.echo(f"Underlying role {underlying_caller_identity['arn']}")
        click.echo("")

    if account_id:
        project['account_id'] = account_id
    else:
        project['account_id'] = caller_identity['account_id']

    if verbose:
        click.echo(f"Using account_id {project['account_id']}")

    ctx.obj = {
        'project_filepath': project_file,
        'role_arn': project.get('role_arn'),
        'github_repository': project.get('github_repository'),
        'tf_stack_root': project.get('tf_stack_root'),
        'verbose': verbose,
        'dry_run': dry_run,
        'project': project,
    }


@cli.command()
@click.option("--service-id", required=True)
@click.option("--namespace", required=True, default=DEFAULT_ECR_NAMESPACE)
@click.option("--label", default="latest")
@click.pass_context
def publish(ctx, service_id, namespace, label):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']
    dry_run = ctx.obj['dry_run']

    account_id = project['account_id']
    region_id = project['aws_region_name']

    ecr = Ecr(account_id, region_id, role_arn)
    parameter_store = SsmParameterStore(project['id'], role_arn)

    print(f"*** Attempting to publish {project['id']}/{service_id}")

    profile_name = None
    if role_arn:
        profile_name = 'service_publisher'
        configure_aws_profile(role_arn, profile_name)

    print(f"*** Authenticating {account_id} for `docker push` with ECR")
    ecr.login(profile_name)

    remote_image_name = ecr.publish_image(
        namespace,
        service_id,
        label,
        dry_run
    )

    parameter_store.update_ssm(
        service_id,
        label,
        remote_image_name,
        dry_run
    )

    print(f"*** Done publishing {project['id']}/{service_id}")


@cli.command()
@click.option('--project-name', '-n', prompt="Enter a descriptive name for this project",
              help="The name of the project")
@click.option('--environment-id', '-e', prompt="Enter an id for an environment", help="The primary environment's ID")
@click.option('--environment-name', '-a', prompt="Enter a descriptive name for this environment",
              help="The primary environment's name")
@click.pass_context
def initialise(ctx, project_name, environment_id, environment_name):
    role_arn = ctx.obj['role_arn']
    verbose = ctx.obj['verbose']
    dry_run = ctx.obj['dry_run']
    project = ctx.obj['project']

    project_filepath = ctx.obj['project_filepath']

    releases_store = DynamoDbReleaseStore(project["id"], role_arn)

    project = {
        'id': project["id"],
        'name': project_name,
        'role_arn': role_arn,
        'environments': [
            {
                'id': environment_id,
                'name': environment_name
            }
        ]
    }

    if verbose:
        click.echo(pprint(project))
    if not dry_run:
        if exists(project_filepath):
            click.confirm(
                f"This will replace existing project file ({project_filepath}), do you want to continue?",
                abort=True)

        click.confirm(f"{releases_store.describe_initialisation()}?")
        releases_store.initialise()

        save(project_filepath, project)
    elif verbose:
        click.echo("dry-run, not created.")


@cli.command()
@click.option('--release-id', prompt="Release ID to deploy", default="latest", show_default=True,
              help="The ID of the release to be deployed, or the latest release if unspecified")
@click.option('--environment-id', prompt="Environment ID to deploy release to",
              help="The target environment of this deployment")
@click.option("--namespace", required=True, default=DEFAULT_ECR_NAMESPACE)
@click.option('--description', prompt="Enter a description for this deployment",
              help="A description of this deployment", default="No description given.")
@click.pass_context
def deploy(ctx, release_id, environment_id, namespace, description):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']
    dry_run = ctx.obj['dry_run']
    verbose = ctx.obj['verbose']

    releases_store = DynamoDbReleaseStore(project['id'], role_arn)
    parameter_store = SsmParameterStore(project['id'], role_arn)
    account_id = project['account_id']
    region_id = project['aws_region_name']

    ecr = Ecr(account_id, region_id, role_arn)

    user_details = Iam(role_arn)

    environments = get_environments_lookup(project)

    try:
        environment = environments[environment_id]
    except KeyError:
        raise ValueError(f"Unknown environment. Expected '{environment_id}' in {environments}")

    if release_id == "latest":
        release = releases_store.get_latest_release()
    else:
        release = releases_store.get_release(release_id)

    click.echo(click.style("Release to deploy:", fg="blue"))
    click.echo(click.style(f"Release ID: {release['release_id']}", fg="green"))
    click.echo(click.style(f"Requested by: {release['requested_by']}", fg="yellow"))
    click.echo(click.style(f"Date created: {release['date_created']}", fg="yellow"))
    click.echo(pprint(release['images']))
    click.confirm(click.style("create deployment?", fg="green", bold=True), abort=True)

    caller_identity = user_details.caller_identity(underlying=True)
    deployment = create_deployment(environment, caller_identity['arn'], description)

    click.echo(click.style("Created deployment:", fg="blue"))
    click.echo(click.style(f"Requested by: {deployment['requested_by']}", fg="yellow"))
    click.echo(click.style(f"Date created: {deployment['date_created']}", fg="yellow"))

    releases_store.add_deployment(
        release_id=release['release_id'],
        deployment=deployment,
        dry_run=dry_run
    )

    for service_id, image_name in release['images'].items():
        ssm_path = parameter_store.update_ssm(
            service_id=service_id,
            label=environment_id,
            image_name=image_name,
            dry_run=dry_run
        )

        if verbose:
            click.echo(f"*** Updated SSM path {ssm_path} to {image_name}")

        old_tag = image_name.split(":")[-1]
        new_tag = environment_id

        ecr.retag_image(
            namespace=namespace,
            service_id=service_id,
            tag=old_tag,
            new_tag=new_tag,
            dry_run=dry_run
        )

        if verbose:
            click.echo(f"*** Retagged image {service_id}:{old_tag} to {service_id}:{new_tag} ")

    if dry_run:
        click.echo("dry-run, not created.")


@cli.command()
@click.option('--from-label', prompt="Label to base release upon",
              help="The existing label upon which this release will be based", default="latest", show_default=True)
@click.option('--service-id', prompt="Service to update", default="all", show_default=True,
              help="The service to update with a (prompted for) new image")
@click.option('--release-description', prompt="Description for this release", default="No description given.")
@click.pass_context
def prepare(ctx, from_label, service_id, release_description):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']
    dry_run = ctx.obj['dry_run']

    releases_store = DynamoDbReleaseStore(project['id'], role_arn)
    parameter_store = SsmParameterStore(project['id'], role_arn)
    user_details = Iam(role_arn)

    from_images = parameter_store.get_services_to_images(from_label)
    service_source = "latest"
    if service_id == "all":
        release_image = {}
    else:
        service_source = click.prompt("Label or image URI to release for specified service", default="latest")
        if _is_url(service_source):
            release_image = {service_id: service_source}
        else:
            release_image = parameter_store.get_service_to_image(service_source, service_id)
    release_images = {**from_images, **release_image}

    if not release_images:
        raise click.UsageError(f"No images found for {project['id']} {service_id} {from_label}")

    caller_identity = user_details.caller_identity(underlying=True)

    release = create_release(
        project['id'],
        project['name'],
        caller_identity['arn'],
        release_description,
        release_images)

    if service_id == "all":
        click.echo(f"Prepared release from images in {from_label}")
    else:
        click.echo(f"Prepared release from images in {from_label} with {service_id} from {service_source}")
    click.echo(pprint(release))

    if not dry_run:
        releases_store.put_release(release)
    else:
        click.echo("dry-run, not created.")


@cli.command()
@click.argument('release_id', required=False)
@click.pass_context
def show_release(ctx, release_id):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']

    releases_store = DynamoDbReleaseStore(project['id'], role_arn)

    if not release_id:
        release = releases_store.get_latest_release()
    else:
        release = releases_store.get_release(release_id)
    click.echo(json.dumps(release, sort_keys=True, indent=2))


@cli.command()
@click.argument('release_id', required=False)
@click.pass_context
def show_deployments(ctx, release_id):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']

    releases_store = DynamoDbReleaseStore(project['id'], role_arn)

    if not release_id:
        releases = releases_store.get_recent_deployments()
    else:
        releases = [releases_store.get_release(release_id)]
    summaries = _summarise_release_deployments(releases)
    for summary in summaries:
        click.echo("{release_id} {environment_id} {deployed_date} '{description}'".format(**summary))


@cli.command()
@click.option('--label', '-l', help="The label to show (e.g., latest')")
@click.pass_context
def show_images(ctx, label):
    project = ctx.obj['project']
    role_arn = ctx.obj['role_arn']
    parameter_store = SsmParameterStore(project['id'], role_arn)

    images = parameter_store.get_images(label=label)

    summaries = sorted(_summarise_ssm_response(images), key=lambda k: k['name'])

    paths = {}
    for summary in summaries:
        result = {}
        ecr_uri = _format_ecr_uri(summary['value'])
        name = summary['name']

        result['image_name'] = "{label:>25}:{tag}".format(**ecr_uri)
        result['last_modified'] = summary['last_modified']

        paths[name] = "{last_modified} {image_name}".format(**result)

    click.echo("\n".join(pprint_path_keyval_dict(paths)))


def main():
    cli()



from .iam import Iam


class SsmParameterStore:
    def __init__(self, project_id, role_arn=None):
        self.project_id = project_id
        self.session = Iam.get_session("ReleaseToolSsmParameterStore", role_arn)
        self.ssm = self.session.client('ssm')

    @staticmethod
    def _image_to_service_name(image):
        return image.rsplit("/")[-1]

    def get_parameters_by_path(self, *args, **kwargs):
        paginator = self.ssm.get_paginator("get_parameters_by_path")

        for page in paginator.paginate(*args, **kwargs):
            yield from page["Parameters"]

    def get_images(self, label=None):
        ssm_path = self.create_ssm_key(label)

        return self.get_parameters_by_path(
            Path=ssm_path,
            Recursive=True
        )

    def _get_parameter(self, path):
        response = self.ssm.get_parameter(Name=path)
        if response['ResponseMetadata']['HTTPStatusCode'] != 200:
            raise ValueError(
                f"SSM get parameter failed {response['ResponseMetadata']}")
        return response['Parameter']

    def get_services_to_images(self, label):
        ssm_parameters = {d["Name"]: d["Value"]
                          for d in self.get_images(label)}

        return {
            SsmParameterStore._image_to_service_name(key): value for key, value in ssm_parameters.items()
        }

    def get_service_to_image(self, label, service):
        image_path = self.create_ssm_key(label, service)
        parameter = self.ssm.get_parameter(Name=image_path)

        return {
            SsmParameterStore._image_to_service_name(parameter["Parameter"]["Name"]): parameter["Parameter"]["Value"]
        }

    def create_ssm_key(self, label=None, service_id=None):
        # https://github.com/wellcomecollection/platform/tree/master/docs/rfcs/013-release_deployment_tracking#build-artefacts-ssm-parameters
        # Keys are referenced with the following paths:
        #   /{project_id}/images/{label}/{service_id}
        ssm_key_parts = filter(
            lambda part: part is not None,
            ['', self.project_id, 'images', label, service_id])
        ssm_key = "/".join(ssm_key_parts)
        return ssm_key

    def update_ssm(self, service_id, label, image_name, dry_run=False):
        ssm_path = self.create_ssm_key(label, service_id)

        if not dry_run:
            self.ssm.put_parameter(
                Name=ssm_path,
                Description=f"Docker image URL; auto-managed by {__file__}",
                Value=image_name,
                Type="String",
                Overwrite=True
            )

        return ssm_path
